PACKAGE_NAME = "gen-vm-image"
REPO_NAME = "gen-vm-image"
GOCD_GROUP = "bare_metal_vm_image"
GOCD_TEMPLATE = "bare_metal_vm_image"
GOCD_FORMAT_VERSION = 10
GO_REVISION_COMMIT_VAR = "GO_REVISION_SIF_VM_IMAGES"
CLOUD_CONFIG_DIR = "cloud-init-config"
IMAGE_CONFIG_DIR = "image-config"
GENERATED_IMAGE_DIR = "generated-images"
VM_DISK_DIR = "vmdisks"
TMP_DIR = "tmp"
